# **************************************************************************** #
#                                                                              #
#                                                         :::      ::::::::    #
#    plot_periods.py                                    :+:      :+:    :+:    #
#                                                     +:+ +:+         +:+      #
#    By: daniloceano <danilo.oceano@gmail.com>      +#+  +:+       +#+         #
#                                                 +#+#+#+#+#+   +#+            #
#    Created: 2024/01/02 23:38:49 by daniloceano       #+#    #+#              #
#    Updated: 2024/07/14 10:10:32 by daniloceano      ###   ########.fr        #
#                                                                              #
# **************************************************************************** #

import logging
import os

import numpy as np
import pandas as pd
import xarray as xr


def plot_periods(
    out_track: pd.DataFrame,
    times: pd.DatetimeIndex,
    lat: xr.DataArray,
    results_subdirectory: str,
    figures_directory: str,
    app_logger: logging.Logger,
    processed_vorticity: bool = False,
):
    """
    Calls the cyclophaser function to determine the life cycle periods.

    Parameters:
        out_track (pd.DataFrame): The output track data.
        times (pd.DatetimeIndex): The time values.
        lat (xr.DataArray): The latitude values.
        app_logger: The application logger.

    Returns:
        The result of the determine_periods function.
    """
    from cyclophaser import determine_periods

    app_logger.info("Calling Cyclophaser for determining life cycle periods...")

    periods_figure_directory = os.path.join(figures_directory, "Periods")
    os.makedirs(periods_figure_directory, exist_ok=True)
    vorticity_data = out_track["min_max_zeta_850"].to_list()

    options_high_res = {
        "use_filter": "auto",
        "replace_endpoints_with_lowpass": 24,
        "use_smoothing": "auto",
        "use_smoothing_twice": False,
        "savgol_polynomial": 3,
        "cutoff_low": 168,
        "cutoff_high": 48,
    }

    options_low_res = {
        "use_filter": len(vorticity_data) // 6 if len(vorticity_data) // 6 > 3 else 4,
        "replace_endpoints_with_lowpass": 24,
        "use_smoothing": (
            len(vorticity_data) // 8 | 1 if len(vorticity_data) // 8 | 1 > 3 else 4
        ),
        "use_smoothing_twice": False,
        "savgol_polynomial": 3,
        "cutoff_low": 168,
        "cutoff_high": 12,
    }

    # Select options based on data resolution
    dx = float(np.abs(lat[1] - lat[0]).values)
    if dx < 2:
        options = options_high_res
        app_logger.info(
            f"Using high resolution options for cyclophaser (dx = {dx}): %s", options
        )
    else:
        options = options_low_res
        app_logger.info(
            f"Using low resolution options for cyclophaser (dx = {dx}): %s", options
        )

    # If vorticity is given and the information is on trackfile, it means that
    # it is already processed
    if processed_vorticity:
        options = {
            "use_filter": False,
            "use_smoothing": "auto",
            "use_smoothing_twice": "auto",
        }
        app_logger.info(
            "Vorticity already processed, using default smoothing options for cyclophaser: %s",
            options,
        )

    periods_args = {
        "plot": os.path.join(periods_figure_directory, "periods"),
        "plot_steps": os.path.join(periods_figure_directory, "periods_steps"),
        "export_dict": os.path.join(results_subdirectory, "periods"),
        "process_vorticity_args": options,
    }

    try:
        determine_periods(vorticity_data, x=times, **periods_args)
    except Exception as e:
        app_logger.error(f"Error calling cyclophaser: {e}")
        return



# **************************************************************************** #
#                                                                              #
#                                                         :::      ::::::::    #
#    timeseries_terms.py                                :+:      :+:    :+:    #
#                                                     +:+ +:+         +:+      #
#    By: daniloceano <danilo.oceano@gmail.com>      +#+  +:+       +#+         #
#                                                 +#+#+#+#+#+   +#+            #
#    Created: 2022/05/22 17:56:57 by daniloceano       #+#    #+#              #
#    Updated: 2023/12/29 14:04:06 by daniloceano      ###   ########.fr        #
#                                                                              #
# **************************************************************************** #

"""
This script reads an CSV file with energy and conversion terms from the Lorenz
Energy Cycle (as input from user) and plot a timeseries for each.

Created by:
    Danilo Couto de Souza
    Universidade de São Paulo (USP)
    Instituto de Astornomia, Ciências Atmosféricas e Geociências
    São Paulo - Brazil

Contact:
    danilo.oceano@gmail.com
"""
import os

import matplotlib.dates as mdates
import matplotlib.pyplot as plt

import src.plots.utils as utils
from src.plots.utils import read_results


def _plotter(df, term_list, figures_directory):
    """Plots a timeseries for given terms."""

    plt.close("all")
    plt.figure(figsize=(10, 8))
    ax = plt.gca()

    terms = utils.TERM_DETAILS[term_list]["terms"]
    title = utils.TERM_DETAILS[term_list]["label"]
    y_label = f'[{utils.TERM_DETAILS[term_list]["unit"]}]'
    times = df.index

    for i, term in enumerate(terms):
        if term in df.columns:
            plt.plot(
                times,
                df[term],
                c=utils.COLORS[i],
                label=term,
                lw=utils.LINEWIDTH,
                marker=utils.MARKERS[i],
                markersize=8,
                markerfacecolor=utils.COLORS[i],
                markeredgecolor=utils.MARKER_EDGE_COLOR,
                markeredgewidth=0.5,
                linestyle=utils.LINESTYLE,
            )

    plt.grid(c="gray", linewidth=0.15, linestyle="-")
    plt.legend(prop={"size": 18})
    plt.title(title, fontsize=18)
    plt.xticks(fontsize=14)
    plt.yticks(fontsize=14)
    plt.ylabel(y_label, fontsize=18)
    plt.xlim(times[0], times[-1])
    plt.tick_params(axis="x", labelrotation=20)
    ax.xaxis.set_major_formatter(mdates.DateFormatter("%m-%d %H"))

    plt.savefig(os.path.join(figures_directory, f"timeseires_{term_list}.png"))


def plot_timeseries(results_file, figures_directory, app_logger=False):
    """Reads data from an results_file and plots timeseries for LEC terms."""
    (
        app_logger.info("Plotting timeseries...")
        if app_logger
        else print("Plotting timeseries...")
    )
    df = read_results(results_file, app_logger)

    figures_subdirectory = os.path.join(figures_directory, "timeseries")
    os.makedirs(figures_subdirectory, exist_ok=True)

    for term_list in utils.TERM_DETAILS:
        _plotter(df, term_list, figures_subdirectory)
        (
            app_logger.info(
                f"Figure saved for {term_list} in directory: {figures_subdirectory}"
            )
            if app_logger
            else print(
                f"Figure saved for {term_list} in directory: {figures_subdirectory}"
            )
        )


if __name__ == "__main__":
    results_file = "samples/sample_results.csv"
    figures_directory = "samples/Figures/"
    plot_timeseries(results_file, figures_directory)



# **************************************************************************** #
#                                                                              #
#                                                         :::      ::::::::    #
#    map_box_limits.py                                  :+:      :+:    :+:    #
#                                                     +:+ +:+         +:+      #
#    By: daniloceano <danilo.oceano@gmail.com>      +#+  +:+       +#+         #
#                                                 +#+#+#+#+#+   +#+            #
#    Created: 2022/06/14 14:58:51 by daniloceano       #+#    #+#              #
#    Updated: 2023/12/29 14:02:22 by daniloceano      ###   ########.fr        #
#                                                                              #
# **************************************************************************** #

"""
Created by:
    Danilo Couto de Souza
    Universidade de São Paulo (USP)
    Instituto de Astornomia, Ciências Atmosféricas e Geociências
    São Paulo - Brazil

Contact:
    danilo.oceano@gmail.com

"""

import os

import cartopy.crs as ccrs
import matplotlib.pyplot as plt
from shapely.geometry.polygon import Polygon

from src.plots.utils import (map_borders, read_box_limits, setup_gridlines,
                             setup_map)


def plot_box_limits(box_limits_file, figures_directory, app_logger=False):
    """
    Plots the box limits on a map.

    Parameters:
        df (pandas.DataFrame): DataFrame containing the box limits.
        figures_directory (str): Directory to save the figure.
        app_logger: Optional logger for logging messages.
    """
    (
        app_logger.info("Plotting box limits on map...")
        if app_logger
        else print("Plotting box limits on map...")
    )
    df = read_box_limits(box_limits_file)

    min_lon, max_lon = df.loc["min_lon"].item(), df.loc["max_lon"].item()
    min_lat, max_lat = df.loc["min_lat"].item(), df.loc["max_lat"].item()

    plt.close("all")
    fig, ax = plt.subplots(
        figsize=(8, 8), subplot_kw={"projection": ccrs.PlateCarree()}
    )
    ax.set_extent(
        [min_lon - 20, max_lon + 20, max_lat + 20, min_lat - 20], crs=ccrs.PlateCarree()
    )
    map_borders(ax)
    setup_map(ax)
    setup_gridlines(ax)

    # Plot selected domain
    pgon = Polygon(
        [
            (min_lon, min_lat),
            (min_lon, max_lat),
            (max_lon, max_lat),
            (max_lon, min_lat),
            (min_lon, min_lat),
        ]
    )
    ax.add_geometries(
        [pgon],
        crs=ccrs.PlateCarree(),
        facecolor="none",
        edgecolor="#BF3D3B",
        linewidth=3,
        alpha=1,
        zorder=3,
    )

    fig.savefig(os.path.join(figures_directory, "box_limits.png"))
    if app_logger:
        app_logger.info("Figure saved in directory: {}".format(figures_directory))
    else:
        print("Figure saved in directory: {}".format(figures_directory))


if __name__ == "__main__":
    box_limits_file = "samples/sample_box_limits"
    figures_directory = "samples/Figures"
    os.makedirs(figures_directory, exist_ok=True)
    plot_box_limits(box_limits_file, figures_directory)



# **************************************************************************** #
#                                                                              #
#                                                         :::      ::::::::    #
#    generation_and_dissipation_terms.py                :+:      :+:    :+:    #
#                                                     +:+ +:+         +:+      #
#    By: daniloceano <danilo.oceano@gmail.com>      +#+  +:+       +#+         #
#                                                 +#+#+#+#+#+   +#+            #
#    Created: 2020/11/26 18:27:40 by daniloceano       #+#    #+#              #
#    Updated: 2024/07/18 00:33:23 by daniloceano      ###   ########.fr        #
#                                                                              #
# **************************************************************************** #

"""
This script defines the GenerationDissipationTerms object. It uses the MetData object as an
input. The built-in functions use the input data to compute the following
partioned energy contents of the Lorenz Energy Cycle.

Created by:
    Danilo Couto de Souza
    Universidade de São Paulo (USP)
    Instituto de Astornomia, Ciências Atmosféricas e Geociências
    São Paulo - Brazil

Contact:
    danilo.oceano@gmail.com
"""

import logging

import numpy as np
from metpy.constants import Cp_d, g
from metpy.units import units

from ..utils.box_data import BoxData
from ..utils.calc_averages import CalcAreaAverage


class GenerationDissipationTerms:
    """
    Class to compute generation and dissipation terms of the Lorenz Energy Cycle.

    Attributes:
        method (str): The computation method used ('fixed', 'track', or 'choose').
        box_obj (BoxData): The BoxData object containing meteorological data.

    Methods:
        calc_gz: Computes Generation of Zonal Available Potential Energy (Gz).
        calc_ge: Computes Generation of Eddy Available Potential Energy (Ge).
        calc_dz: Computes Dissipation of Zonal Kinetic Energy (Dz).
        calc_de: Computes Dissipation of Eddy Kinetic Energy (De).

    Source for formulas used here:
        Brennan, F. E., & Vincent, D. G. (1980).
        Zonal and Eddy Components of the Synoptic-Scale Energy Budget
        during Intensification of Hurricane Carmen (1974),
        Monthly Weather Review, 108(7), 954-965. Retrieved Jan 25, 2022, from:
        https://journals.ametsoc.org/view/journals/mwre/108/7/1520-0493_1980_108_0954_zaecot_2_0_co_2.xml
    """

    def __init__(self, box_obj: BoxData, method: str, app_logger: logging.Logger):
        """Initialize the EnergyContents object with a BoxData object and a method."""
        self._initialize_attributes(box_obj, method, app_logger)

    def _initialize_attributes(self, box_obj, method, app_logger):
        """Helper method to initialize attributes from the BoxData object."""

        # Operational attributes
        self.method = method
        self.box_obj = box_obj
        self.results_subdirectory = box_obj.results_subdirectory
        self.results_subdirectory_vertical_levels = box_obj.results_subdirectory_vertical_levels
        self.app_logger = app_logger

        # Initialize spatial and temporal attributes
        self.LonIndexer = box_obj.LonIndexer
        self.LatIndexer = box_obj.LatIndexer
        self.VerticalCoordIndexer = box_obj.VerticalCoordIndexer
        self.TimeName = box_obj.TimeName

        # Initialize lengths for averaging
        self.xlength = box_obj.xlength
        self.ylength = box_obj.ylength
        self.method = method
        self.box_obj = box_obj

        # Initialize pressure data
        self.PressureData = box_obj.PressureData

        # Initialize attributes related to temperature
        self.tair = box_obj.tair
        self.tair_AE = box_obj.tair_AE
        self.tair_ZE = box_obj.tair_ZE

        # Initialize attributes related to wind components
        self.u = box_obj.u
        self.u_ZA = box_obj.u_ZA
        self.u_ZE = box_obj.u_ZE
        self.v = box_obj.v
        self.v_ZA = box_obj.v_ZA
        self.v_ZE = box_obj.v_ZE

        # Initialize attributes related to wind stress
        self.ust_ZA = box_obj.ust_ZA
        self.ust_ZE = box_obj.ust_ZE
        self.vst_ZE = box_obj.vst_ZE
        self.vst_ZA = box_obj.vst_ZA

        # Initialize attributes related to vertical velocity
        self.omega = box_obj.omega

        # Initialize attributes related to static stability
        self.sigma_AA = box_obj.sigma_AA
        self.sigma_AA = box_obj.sigma_AA

        # Initialize attributes related to the adiabatic heating term
        self.Q = box_obj.Q
        self.Q_ZA = box_obj.Q_ZA
        self.Q_AA = box_obj.Q_AA
        self.Q_ZE = box_obj.Q_ZE
        self.Q_AE = box_obj.Q_AE

    def calc_gz(self):
        """Computes Generation of Zonal Available Potential Energy (Gz)."""
        self.app_logger.debug("Computing Gz...")
        term = self.Q_AE * self.tair_AE
        function = CalcAreaAverage(term, self.ylength) / (Cp_d * self.sigma_AA)
        function = self._handle_nans(function)
        self._save_vertical_levels(function, "Gz")
        Gz = (
            function.integrate(self.VerticalCoordIndexer)
            * self.PressureData.metpy.units
        )
        self._convert_units(Gz, "Gz")
        self.app_logger.debug("Ok.")
        return Gz

    def calc_ge(self):
        """Computes Generation of Eddy Available Potential Energy (Ge)."""
        self.app_logger.debug("Computing Ge...")
        term = self.Q_ZE * self.tair_ZE
        function = CalcAreaAverage(term, self.ylength, xlength=self.xlength) / (
            Cp_d * self.sigma_AA
        )
        function = self._handle_nans(function)
        self._save_vertical_levels(function, "Ge")
        Ge = (
            function.integrate(self.VerticalCoordIndexer)
            * self.PressureData.metpy.units
        )
        self._convert_units(Ge, "Ge")
        self.app_logger.debug("Ok.")
        return Ge

    def calc_dz(self):
        """
        Computes Dissipation of Zonal Kinetic Energy (Dz).

        Note: Still needs to be fully implemented and tested
        """
        self.app_logger.debug("Computing Dz...")
        # Here we will use only the lowest vertical level
        term = (self.u_ZA.isel({self.VerticalCoordIndexer: 0}) * self.ust_ZA) + (
            self.v_ZA.isel({self.VerticalCoordIndexer: 0}) * self.vst_ZA
        )
        function = CalcAreaAverage(term, self.ylength) / g
        self._save_vertical_levels(function, "Dz")
        Dz = units.Pa * function
        self._convert_units(Dz, "Dz")
        self.app_logger.debug("Ok.")
        return Dz

    def calc_de(self):
        """
        Computes Dissipation of Eddy Kinetic Energy (De).

        Note: Still needs to be fully implemented and tested
        """
        self.app_logger.debug("Computing De...")
        # Here we will use only the lowest vertical level
        term = (self.u_ZE.isel({self.VerticalCoordIndexer: 0}) * self.ust_ZE) + (
            self.v_ZE.isel({self.VerticalCoordIndexer: 0}) * self.vst_ZE
        )
        function = CalcAreaAverage(term, self.ylength) / g
        self._save_vertical_levels(function, "De")
        De = units.Pa * function
        self._convert_units(De, "De")
        self.app_logger.debug("Ok.")
        return De

    def _handle_nans(self, function):
        """
        If there are any, interpolate them and drop any remaining NaN values.
        If there are still NaN values after interpolation, drop the dimensions that contain NaN values.

        Parameters:
            function (np.ndarray): The function to handle NaN values for.

        Returns:
            None
        """
        if np.isnan(function).any():
            function = (
                function.interpolate_na(dim=self.VerticalCoordIndexer)
                * function.metpy.units
            )
            if np.isnan(function).any():
                function = function.dropna(dim=self.VerticalCoordIndexer)
        return function

    def _convert_units(self, function, variable_name):
        """
        Converts the units of a given function to 'J/m^2'.

        Parameters:
            function (type): The function to be converted.
            variable_name (type): The name of the variable associated with the function.

        Raises:
            ValueError: If there is a unit error in the given variable.

        Returns:
            None
        """
        try:
            function = function.metpy.convert_units("W/m^2")
        except ValueError as e:
            raise ValueError(f"Unit error in {variable_name}") from e
        return function

    def _save_vertical_levels(self, function, variable_name):
        """Save computed energy data to a CSV file."""
        df = function.to_dataframe(name=variable_name)
        df.reset_index(inplace=True)
        if self.method == "fixed":
            df = df.pivot(index=self.TimeName, columns=self.VerticalCoordIndexer)
        else:
            df.set_index(self.TimeName, inplace=True)
            df.index = df.index.strftime("%Y-%m-%d %H:%M:%S")
            df = df.pivot(columns=self.VerticalCoordIndexer, values=variable_name)
            df.columns.name = None

        df.to_csv(
            f"{self.results_subdirectory_vertical_levels}/{variable_name}_{self.VerticalCoordIndexer}.csv",
            mode="a",
            header=None,
        )



# **************************************************************************** #
#                                                                              #
#                                                         :::      ::::::::    #
#    tools.py                                           :+:      :+:    :+:    #
#                                                     +:+ +:+         +:+      #
#    By: daniloceano <danilo.oceano@gmail.com>      +#+  +:+       +#+         #
#                                                 +#+#+#+#+#+   +#+            #
#    Created: 2023/12/19 17:33:03 by daniloceano       #+#    #+#              #
#    Updated: 2024/07/18 09:38:56 by daniloceano      ###   ########.fr        #
#                                                                              #
# **************************************************************************** #

import argparse
import logging
import os
from datetime import timedelta

import dask
import numpy as np
import pandas as pd
import xarray as xr
from metpy.units import units

from .select_area import slice_domain


def initialize_logging(results_subdirectory, args):
    """
    Initializes the logging configuration for the application.

    Args:
        results_subdirectory (str): Directory path to save the log file.
        args (object): The argparse object containing the command line arguments.
    """

    verbose = True if args.verbosity else False

    # Set root logger to higher severity level (INFO or ERROR)
    root_log_level = logging.ERROR if not verbose else logging.INFO
    logging.basicConfig(
        level=root_log_level, format="%(asctime)s - %(levelname)s - %(message)s"
    )

    # Create a separate logger for the application
    app_logger = logging.getLogger("lorenzcycletoolkit")
    app_log_level = logging.DEBUG if verbose else logging.INFO
    app_logger.setLevel(app_log_level)
    # Prevent the logger from propagating messages to the root logger
    app_logger.propagate = False

    # Create file handler for saving logs
    log_file_name = f'log.{os.path.basename(args.infile).split(".")[0]}'
    log_file = os.path.join(results_subdirectory, log_file_name)
    file_handler = logging.FileHandler(log_file, mode="w")
    file_handler.setLevel(app_log_level)
    file_formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    file_handler.setFormatter(file_formatter)
    app_logger.addHandler(file_handler)

    # Create a console handler for app logger
    console_handler = logging.StreamHandler()
    console_handler.setLevel(app_log_level)
    console_handler.setFormatter(file_formatter)
    app_logger.addHandler(console_handler)

    return app_logger


def convert_longitude_range(df: xr.Dataset, lon_indexer: str) -> xr.Dataset:
    """
    Convert longitude range from 0-360 degrees to -180 to 180 degrees.

    This function modifies the longitude coordinates of the provided xarray Dataset
    and sorts the data based on these updated longitudes.

    Args:
        df (xr.Dataset): The dataset containing longitude coordinates.
        lon_indexer (str): The name of the longitude coordinate in the dataset.

    Returns:
        xr.Dataset: The dataset with updated longitude coordinates.
    """
    df.coords[lon_indexer] = (df.coords[lon_indexer] + 180) % 360 - 180
    df = df.sortby(df[lon_indexer])
    return df


def find_extremum_coordinates(
    data: xr.DataArray, lat: xr.DataArray, lon: xr.DataArray, variable: str
) -> tuple:
    """
    Finds the indices of extremum values for a given variable in a dataset.

    Args:
        data (xr.DataArray): DataArray containing the data for calculation.
        lat (xr.DataArray): DataArray containing the latitudes.
        lon (xr.DataArray): DataArray containing the longitudes.
        variable (str): Name of the variable to find extremum indices.

    Returns:
        tuple: Tuple containing latitude and longitude of the extremum value.
    """
    lat_values = lat.values
    lon_values = lon.values
    min_lat = lat_values.min()

    data = np.array(data)  # Ensure data is a numpy array

    if variable == "min_max_zeta":
        index = np.unravel_index(
            data.argmin() if min_lat < 0 else data.argmax(), data.shape
        )
    elif variable in ["min_hgt", "max_wind"]:
        index = np.unravel_index(
            data.argmin() if variable == "min_hgt" else data.argmax(), data.shape
        )
    else:
        logging.error(f"Invalid variable specified: {variable}")
        raise ValueError(f"Invalid variable specified: {variable}")

    return lat_values[index[0]], lon_values[index[1]]


def get_cdsapi_data(
    args: argparse.Namespace, track: pd.DataFrame, app_logger: logging.Logger
) -> xr.Dataset:
    import math

    import cdsapi

    # Extract bounding box (lat/lon limits) from track
    min_lat, max_lat = track["Lat"].min(), track["Lat"].max()
    min_lon, max_lon = track["Lon"].min(), track["Lon"].max()

    # Apply a 15-degree buffer and round to nearest integer
    buffered_max_lat = math.ceil(max_lat + 15)
    buffered_min_lon = math.floor(min_lon - 15)
    buffered_min_lat = math.floor(min_lat - 15)
    buffered_max_lon = math.ceil(max_lon + 15)

    # Define the area for the request
    # North, West, South, East. Nort/West/Sout/East
    area = (
        f"{buffered_max_lat}/{buffered_min_lon}/{buffered_min_lat}/{buffered_max_lon}"
    )

    pressure_levels = [
        "1",
        "2",
        "3",
        "5",
        "7",
        "10",
        "20",
        "30",
        "50",
        "70",
        "100",
        "125",
        "150",
        "175",
        "200",
        "225",
        "250",
        "300",
        "350",
        "400",
        "450",
        "500",
        "550",
        "600",
        "650",
        "700",
        "750",
        "775",
        "800",
        "825",
        "850",
        "875",
        "900",
        "925",
        "950",
        "975",
        "1000",
    ]

    variables = [
        "u_component_of_wind",
        "v_component_of_wind",
        "temperature",
        "vertical_velocity",
        "geopotential",
    ]

    # Assuming your data resolution is 3 hours and your service updates at 00,
    # 03, 06, ..., 21 hours

    # Convert track index to DatetimeIndex and find the last date & time
    track_datetime_index = pd.DatetimeIndex(track.index)
    last_track_timestamp = track_datetime_index.max()

    # Calculate if the additional day is needed by comparing last track
    # timestamp with the last possible data timestamp for that day
    last_possible_data_timestamp_for_day = pd.Timestamp(
        f"{last_track_timestamp.strftime('%Y-%m-%d')} 21:00:00"
    )
    need_additional_day = last_track_timestamp > last_possible_data_timestamp_for_day

    # Include additional day in dates if needed
    dates = track_datetime_index.strftime("%Y%m%d").unique()
    if need_additional_day:
        additional_day = (last_track_timestamp + timedelta(days=1)).strftime("%Y%m%d")
        dates = np.append(dates, additional_day)

    # Convert unique dates to string format for the request
    # dates = track.index.strftime('%Y%m%d').unique().tolist()
    time_range = f"{dates[0]}/{dates[-1]}"
    time_step = str(int((track.index[1] - track.index[0]).total_seconds() / 3600))
    time_step = "3" if time_step < "3" else time_step

    # Log track file bounds and requested data bounds
    app_logger.debug(
        f"Track File Limits: max_lon (east): min_lon (west): {min_lon}, "
        f"max_lon (west): {max_lon}, min_lat (south): {min_lat}, "
        f"max_lat (north): {max_lat}"
    )
    app_logger.debug(
        f"Buffered Data Bounds: min_lon (west): {buffered_min_lon}, "
        f"max_lon (east): {buffered_max_lon}, min_lat (south): {buffered_min_lat}, "
        f"max_lat (north): {buffered_max_lat}"
    )
    app_logger.debug(
        f"Requesting data for time range: {time_range}, and time step: {time_step}..."
    )

    # Load ERA5 data
    app_logger.info("Retrieving data from CDS API...")
    c = cdsapi.Client(timeout=600)
    c.retrieve(
        "reanalysis-era5-pressure-levels",
        {
            "product_type": "reanalysis",
            "format": "netcdf",
            "pressure_level": pressure_levels,
            "date": time_range,
            "area": area,
            "time": f"00/to/23/by/{time_step}",
            "variable": variables,
        },
        args.infile,  # save file as passed in arguments
    )

    if not os.path.exists(args.infile):
        raise FileNotFoundError("CDS API file not created.")
    return args.infile


def get_data(args: argparse.Namespace, app_logger: logging.Logger) -> xr.Dataset:
    """
    Opens a NetCDF file and extracts variables specified in a CSV file.

    Args:
        args (argparse.Namespace): Command-line arguments.
        variable_list_df (pd.DataFrame): DataFrame containing the variables to extract.
        app_logger (logging.Logger): Logger for the application.

    Returns:
        xr.Dataset: Dataset containing extracted variables.

    Raises:
        FileNotFoundError: If CSV file or NetCDF file is not found.
        Exception: For other errors occurring during file opening.
    """
    infile = args.infile

    if args.cdsapi:
        if not os.path.exists(infile):
            app_logger.debug(
                "CDS API data not found. Attempting to retrieve data from CDS API..."
            )
            track = pd.read_csv(
                args.trackfile, parse_dates=[0], delimiter=";", index_col="time"
            )
            infile = get_cdsapi_data(args, track, app_logger)
            app_logger.debug(f"CDS API data: {infile}")
        else:
            app_logger.info("CDS API data found.")

    app_logger.debug("Opening input data... ")
    try:
        with dask.config.set(array={"slicing": {"split_large_chunks": True}}):
            data = xr.open_dataset(infile)
    except FileNotFoundError:
        app_logger.error(
            "Could not open file. Check if path, namelist file, and file format (.nc) are correct."
        )
        raise
    except Exception as e:
        app_logger.exception("An exception occurred: {}".format(e))
        raise
    app_logger.debug("Ok.")

    return data


def process_data(
    data: xr.Dataset,
    args: argparse.Namespace,
    variable_list_df: pd.DataFrame,
    app_logger: logging.Logger,
) -> xr.Dataset:
    """
    Process the given data and return a modified dataset.

    Parameters:
    - data: A dataset containing the data to be processed (type: xr.Dataset).
    - args: An argparse.Namespace object containing the command line arguments (type: argparse.Namespace).
    - variable_list_df: A DataFrame containing a list of variables (type: pd.DataFrame).
    - app_logger: A logger object for logging debug messages (type: logging.Logger).

    Returns:
    - data: A modified dataset after processing (type: xr.Dataset).
    """
    # Select only data matching the track dates
    if args.track:
        app_logger.debug("Selecting only data matching the track dates... ")
        track_file = args.trackfile
        track = pd.read_csv(
            track_file, parse_dates=[0], delimiter=";", index_col="time"
        )
        TimeIndexer = variable_list_df.loc["Time"]["Variable"]
        # If using CDS API, resample track to data time step
        if args.cdsapi:
            time_delta = int(
                (data[TimeIndexer][1] - data[TimeIndexer][0]) / np.timedelta64(1, "h")
            )
            track = track[track.index.hour % time_delta == 0]
        data = data.sel({TimeIndexer: track.index.values})

    if (
        data[variable_list_df.loc["Longitude"]["Variable"]].min() < -180
        or data[variable_list_df.loc["Longitude"]["Variable"]].max() > 180
    ):
        data = convert_longitude_range(
            data, variable_list_df.loc["Longitude"]["Variable"]
        )

    LonIndexer = variable_list_df.loc["Longitude"]["Variable"]
    LatIndexer = variable_list_df.loc["Latitude"]["Variable"]
    LevelIndexer = variable_list_df.loc["Vertical Level"]["Variable"]

    app_logger.debug("Assigning geospatial coordinates in radians... ")
    data = data.assign_coords({"rlats": np.deg2rad(data[LatIndexer])})
    data = data.assign_coords({"coslats": np.cos(np.deg2rad(data[LatIndexer]))})
    data = data.assign_coords({"rlons": np.deg2rad(data[LonIndexer])})
    app_logger.debug("Ok.")

    levels_Pa = (
        data[LevelIndexer] * units(str(data[LevelIndexer].units))
    ).metpy.convert_units("Pa")
    data = data.assign_coords({LevelIndexer: levels_Pa})

    data = (
        data.sortby(LonIndexer)
        .sortby(LevelIndexer, ascending=True)
        .sortby(LatIndexer, ascending=True)
    )

    lowest_level = float(data[LevelIndexer].max())
    data = data.sel({LevelIndexer: slice(1000, lowest_level)})

    if args.mpas:
        data = data.drop_dims("standard_height")

    app_logger.debug("Data opened successfully.")
    return data


def prepare_data(
    args, varlist: str = "inputs/namelist", app_logger: logging.Logger = None
) -> xr.Dataset:
    """
    Prepare the data for further analysis.

    Parameters:
        args (object): The arguments for the function.
        varlist (str): The file path to the variable list file (namelist).
        app_logger (logging.Logger): The logger for the application.

    Returns:
        method (str): The method used for the analysis: fixed, track or choose.
        xr.Dataset: The prepared dataset for analysis.
    """

    app_logger.debug(f"Variables specified by the user in: {varlist}")
    app_logger.debug(f"Attempting to read {varlist} file...")
    try:
        variable_list_df = pd.read_csv(varlist, sep=";", index_col=0, header=0)
    except FileNotFoundError:
        app_logger.error("The 'namelist' text file could not be found.")
        raise
    except pd.errors.EmptyDataError:
        app_logger.error("The 'namelist' text file is empty.")
        raise
    app_logger.debug("List of variables found:\n" + str(variable_list_df))

    data = get_data(args, app_logger)

    # Check if variable_list_df matches the data
    if not set(variable_list_df["Variable"]).issubset(set(data.variables)):
        app_logger.error(
            "The variable list does not match the data. Check if the 'namelist' text file is correct."
        )
        raise ValueError("'namelist' text file does not match the data.")

    processed_data = process_data(data, args, variable_list_df, app_logger)
    sliced_data = slice_domain(processed_data, args, variable_list_df)
    return sliced_data



# **************************************************************************** #
#                                                                              #
#                                                         :::      ::::::::    #
#    timeseries_zeta_and_Z.py                           :+:      :+:    :+:    #
#                                                     +:+ +:+         +:+      #
#    By: daniloceano <danilo.oceano@gmail.com>      +#+  +:+       +#+         #
#                                                 +#+#+#+#+#+   +#+            #
#    Created: 2023/12/28 14:48:27 by daniloceano       #+#    #+#              #
#    Updated: 2023/12/29 13:54:38 by daniloceano      ###   ########.fr        #
#                                                                              #
# **************************************************************************** #

import os

import matplotlib.dates as mdates
import matplotlib.pyplot as plt
import pandas as pd

COLORS = ["#BF3D3B", "#023e8a"]
LINEWIDTH = 3


def plot_min_zeta_hgt(track_file, figures_directory, app_logger=False):
    """
    Plots minimum vorticity and geopotential height from a tracking file.

    Parameters:
    track_file (str): Path to the CSV file containing track data.
    figures_directory (str): Directory path to save the plot.
    """
    try:
        track = pd.read_csv(
            track_file, parse_dates=[0], delimiter=";", index_col="time"
        )
    except FileNotFoundError:
        app_logger.info("Error: Track file was not found.") if app_logger else None
        return
    except pd.errors.ParserError:
        (
            app_logger.info("Error: Parsing CSV file failed. Check the file format.")
            if app_logger
            else None
        )
        return

    plt.close("all")
    fig, ax1 = plt.subplots(figsize=(10, 8))

    ax1.plot(
        track.index,
        track["min_max_zeta_850"],
        "--",
        c=COLORS[0],
        lw=LINEWIDTH,
        label=r"$\zeta$",
    )
    ax2 = ax1.twinx()
    ax2.plot(
        track.index, track["min_hgt_850"], "-", c=COLORS[1], lw=LINEWIDTH, label=r"$Z$"
    )

    ax1.set_ylabel("Vorticity [s⁻¹]", fontsize=16, color=COLORS[0])
    ax2.set_ylabel("Geopotential height [m]", fontsize=16, color=COLORS[1])
    ax1.tick_params(axis="x", labelrotation=20, labelsize=14)
    ax1.tick_params(axis="y", colors=COLORS[0], labelsize=14)
    ax2.tick_params(axis="y", colors=COLORS[1], labelsize=14)
    ax1.xaxis.set_major_formatter(mdates.DateFormatter("%m-%d %H"))

    fig.legend(
        loc="upper right",
        bbox_to_anchor=(1, 1),
        bbox_transform=ax1.transAxes,
        fontsize=16,
    )
    ax1.grid(c="gray", linewidth=0.25, linestyle="dashdot", axis="x")
    fig.suptitle(
        r"Central Vorticity ($\zeta$) and Geopotential Height (Z) at 850 hPa",
        fontsize=18,
    )

    # Check if the directory exists and create it if not
    os.makedirs(figures_directory, exist_ok=True)
    plt.savefig(
        os.path.join(figures_directory, "timeseries-min_zeta_hgt.png"),
        bbox_inches="tight",
    )


if __name__ == "__main__":
    track_file = "samples/sample_track.csv"
    figures_directory = "samples/Figures/"
    os.makedirs(figures_directory, exist_ok=True)
    plot_min_zeta_hgt(track_file, figures_directory)
